@@ -12,9 +12,13 @@ tf.get_logger().setLevel("ERROR")
 from classy import data
 from classy import decision_tree
 from classy import defs
-from classy.logging import logger
+from classy.log import logger
 from classy import mixnorm
 from classy import plotting
+from classy import preprocessing
+from classy import tools
+from classy import taxonomies
+from classy import sources
 
 
 class Spectrum:
@@ -27,8 +31,7 @@ class Spectrum:
         pV=None,
         pV_err=None,
         name=None,
-        asteroid_name=None,
-        asteroid_number=None,
+        number=None,
         preprocessed=False,
         **kwargs,
     ):
@@ -50,10 +53,8 @@ class Spectrum:
         pV_err : float
             The albedo uncertainty value. Default is None.
         name : str
-            A unique identifier for this spectrum.
-        asteroid_name : str
             The name of the asteroid the spectrum is referring to.
-        asteroid_number : int
+        number : int
             The number of the asteroid the spectrum is referring to.
         preprocessed : bool
             Whether the reflectance bins and albedo have already been preprocessed
@@ -80,90 +81,68 @@ class Spectrum:
 
         # Assign metadata
         self.name = name
-        self.asteroid_name = asteroid_name
-        self.asteroid_number = asteroid_number
+        self.number = number
 
         # Look up name, number, albedo, if not provided
-        if (asteroid_name is None and asteroid_number is not None) or (
-            asteroid_name is not None and asteroid_number is None
-        ):
-            self.asteroid_name, self.asteroid_number = rocks.id(asteroid_number)
+        if name is None and number is not None:
+            self.name, self.number = rocks.id(number)
+        elif name is not None and number is None:
+            self.name, self.number = rocks.id(name)
+
+        self._refl_original = self.refl.copy()
+        self._wave_original = self.wave.copy()
 
         # Look up pV if it is not provided and we know the asteroid
-        if self.pV is None and self.asteroid_name is not None:
-            rock = rocks.Rock(self.asteroid_name)
+        if self.pV is None and self.name is not None:
+            rock = rocks.Rock(self.name)
             self.pV = rock.albedo.value
             self.pV_err = rock.albedo.error_
         elif self.pV is None:
             self.pV = np.nan
             self.pV_err = np.nan
 
-        # Has it been preprocessed already?
-        self.preprocessed = preprocessed
-
-        if self.preprocessed:
-            self.wave_preprocessed = self.wave
-            self.refl_preprocessed = self.refl
-            self.pV_preprocessed = self.pV
-
         # Assign arbitrary arguments
         self.__dict__.update(**kwargs)
 
+        # Attribute to differentiate user-provided and online spectra
+        self._source = "User"
+
     def __len__(self):
         return len(self.wave)
 
-    def smooth(self, degree=None, window=None, interactive=False):
-        """Smooth spectrum using a Savitzky-Golay filter."""
-        # from scipy.interpolate import UnivariateSpline
-        #
-        # transform = {0: 1, 1: 0.5, 2: 0}
-        #
-        # weights = [transform[f] for f in self.flag]
-        #
-        # if len(self) > 40:
-        #     s = 0.2
-        # else:
-        #     s = 0.2
-        #
-        # spline = UnivariateSpline(x=self.wave, y=self.refl, w=weights, k=5, s=s)
-        # self.refl_smoothed = spline(self.wave)
+    def smooth(self, method="savgol", **kwargs):
+        """Smooth spectrum using a Savitzky-Golay filter or univariate spline.
 
-        if degree is None:
-            degree = 3
-        if window is None:
-            window = int(len(self) // 5)
+        Parameters
+        ----------
+        method : str
+            The smoothing method. Choose from ['savgol', 'spline']. Default is 'savgol'.
+        """
 
-            if window % 2 == 0:
-                window += 1
-            window = window if window > 5 else 5
-
-        if not interactive and degree is None and window is None:
-            logger.warning(
-                "Smoothing is set to non-interactive but no smoothing parameters are set. Running smoothing interactively."
-            )
-            interactive = True
+        if method == "savgol":
+            self.refl = preprocessing.savitzky_golay(self.refl, **kwargs)
+        elif method == "spline":
+            self.refl = preprocessing.univariate_spline(self.wave, self.refl, **kwargs)
         else:
-            self.smooth_degree = degree
-            self.smooth_window = window
-
-        # user might have set refl values to nan
-        self.wave_smoothed = self.wave[~np.isnan(self.refl)]
-        self.refl_smoothed = self.refl[~np.isnan(self.refl)]
-
-        if interactive:
-            # Run with default parameters to initialize self.refl_smoothed
-            self.smooth_window = self.smooth_window if window is not None else 99
-            self.smooth_degree = self.smooth_degre if degree is not None else 3
-            self.refl_smoothed = signal.savgol_filter(
-                self.refl_smoothed, self.smooth_window, self.smooth_degree
+            raise ValueError(
+                f"Unknown smoothing method '{method}'. Choose from ['savgol', 'spline']."
             )
-            self._smooth_interactive()
 
-        # self.refl_original = self.refl
+    @property
+    def albedo(self):
+        return self.pV
 
-        self.refl_smoothed = signal.savgol_filter(
-            self.refl_smoothed, self.smooth_window, self.smooth_degree
-        )
+    @albedo.setter
+    def albedo(self, value):
+        self.pV = value
+
+    @property
+    def albedo_err(self):
+        return self.pV_err
+
+    @albedo_err.setter
+    def albedo_err(self, value):
+        self.pV_err = value
 
     def truncate(self, wave_min, wave_max):
         """Truncate wavelength range to minimum and maximum value.
@@ -181,138 +160,181 @@ class Spectrum:
         if self.wave.size == 0:
             logger.error("No wavelength bins left in spectrum after truncating.")
 
-    def normalize(self):
-        """Normalize the reflectance using the mixnorm algorithm."""
-        alpha = mixnorm.normalize(self)
-        self.refl_preprocessed = np.log(self.refl_normalised) - alpha
-        self.alpha = alpha
-
-    def preprocess(self, smooth_degree=None, smooth_window=None):
-        """Preprocess spectrum for classification.
+    def normalize(self, method="wave", at=None):
+        """Normalize the spectrum.
 
         Parameters
         ----------
-        smooth_degree : int
-            Optional. Degree of the smoothing function applied in the Savtizky-Golay filter.
-            Default is None, which sets the degree to 3.
-        smooth_window : int
-            Optional. Size of data point window in data points applied in the Savtizky-Golay filter.
-            Default is None, which sets the window to a fifth of the number of data points.
+        method : str
+            The method to use for the normalization. Choose from ["wave", "l2", "mixnorm"].
+            Default is "wave".
+        at : float
+            The wavelength at which to normalize. Only relevant when method == "wave".
         """
-        self.detect_features()
-        self.smooth(degree=smooth_degree, window=smooth_window)
-        self.resample()
-        self.normalize()
+        if at is not None:
+            self.refl = preprocessing._normalize_at(self.wave, self.refl, at)
 
-        self.pV_preprocessed = np.log10(self.pV)
-
-        self.preprocessed = True
-
-    def resample(self):
-        """Resample the spectrum to the classy wavelength grid."""
-
-        # Little hack: If the first point is exactly 0.45 or the last point
-        # is exactly 2.45, the spline regards it as extrapolation.
-        # Move it slightly outside to get a value there
-        if self.wave[0] == defs.LIMIT_VIS:
-            self.wave[0] = defs.LIMIT_VIS - 0.0001
-        if self.wave[-1] == defs.LIMIT_NIR:
-            self.wave[-1] = defs.LIMIT_NIR + 0.0001
-
-        if self.refl_smoothed is None:
-            logger.warning(
-                "Interpolating the original spectrum, consider smoothing it first."
-            )
-
-        refl_interp = interpolate.interp1d(
-            self.wave_smoothed, self.refl_smoothed, bounds_error=False
-        )
-
-        # Update basic properties
-        self.wave_interp = defs.WAVE_GRID
-        self.refl_interp = refl_interp(defs.WAVE_GRID)
-        self.mask = np.array([np.isfinite(r) for r in self.refl_interp])
-
-    def classify(self, system="Mahlke+ 2022"):
-        # Find out which system
-        if "mahlke" in system.lower():
-            system = "Mahlke+ 2022"
-        elif "demeo" in system.lower():
-            system = "DeMeo+ 2009"
-        elif "bus" in system.lower():
-            system = "Bus and Binzel"
-        elif "demeo" in system.lower():
-            system = "Tholen 1984"
-
-        if (
-            system == "Mahlke+ 2022"
-            and self.wave.min() >= 2.45
-            or self.wave.max() <= 0.45
-        ):
-            logger.info(
-                f"{self.name}:  Cannot classify following Mahlke+ 2022 - insufficient wavelength coverage."
-            )
-            self.class_mahlke = ""
+            if hasattr(self, "_refl_original"):
+                self._refl_original = preprocessing._normalize_at(
+                    self._wave_original, self._refl_original, at
+                )
             return
 
-        if system == "Mahlke+ 2022" and not self.preprocessed:
-            self.preprocess()
+        if method == "l2":
+            self.refl = preprocessing._normalize_l2(self.refl)
 
-        # Instantiate MCFA model instance if not done yet
-        model = data.load("mcfa")
+            if hasattr(self, "_refl_original"):
+                self._refl_original = preprocessing._normalize_l2(self._refl_original)
 
-        # Get only the classification columns
-        data_input = np.concatenate([self.refl_preprocessed, [self.pV_preprocessed]])[
-            :, np.newaxis
-        ].T
+        elif method == "mixnorm":
+            alpha = mixnorm.normalize(self)
+            self.refl = np.log(self.refl) - alpha
+            self.alpha = alpha
 
-        input_data = pd.DataFrame(
-            {col: val for col, val in zip(defs.COLUMNS["all"], data_input[0])},
-            index=[0],
-        )
+            if hasattr(self, "_refl_original"):
+                self._refl_original = np.log(self._refl_original) - alpha
 
-        # Compute responsibility matrix based on observed values only
-        self.responsibility = model.predict_proba(data_input)
+    #
+    # def preprocess(
+    #     self,
+    #     smooth_method="savgol",
+    #     smooth_params=None,
+    #     resample_params=None,
+    #     taxonomy="mahlke",
+    # ):
+    #     """Preprocess a spectrum for classification in a given taxonomic system.
+    #
+    #     Parameters
+    #     ----------
+    #     smooth_method : str or None
+    #         Optional. The smoothing method to apply to the spectrum. Choose
+    #         from ['savgol', 'spline']. Default is 'savgol'. If set to None, no
+    #         smoothing is applied.
+    #     smooth_params : dict
+    #         Optional. The smoothing parameters passed to the respective
+    #         smoothing functions. Must be valid arguments of the
+    #         ``scipy.signal.savgol_filter`` or ``scipy.interpolate.UnivariateSpline``
+    #         functions depending on the chosen smoothing method.
+    #     resample_params : dict
+    #         Optional. The resampling parameters passed to the ``scipy.interpolate.interp1d`` function.
+    #     taxonomy : str
+    #         Optional. The taxonomic system to prepare the spectrum for. Choose from
+    #         ['mahlke', 'demeo', 'bus', 'tholen']. Default is 'mahlke'.
+    #     """
+    #
+    #     # ------
+    #     # Smoothing is universal and the first step
+    #     if smooth_method == "savgol":
+    #         self.refl = preprocessing.savitzky_golay(self.refl, smooth_params)
+    #         self.refl_plot = self.refl
+    #     elif smooth_method == "spline":
+    #         self.refl = preprocessing.univariate_spline(
+    #             self.wave, self.refl, smooth_params
+    #         )
+    #         self.refl_plot = self.refl
+    #     else:
+    #         # No smoothing
+    #         self.refl_plot = self.refl  # for plotting only
+    #         self.refl = self.refl  # for classification only
+    #
+    #     # Never change the original wave and refl attributes
+    #     # Only change the _pre suffixed ones
+    #     self.wave_pre = self.wave
+    #     self.wave_plot = self.wave
+    #
+    #     # ------
+    #     # Remaining steps depend on the taxonomic scheme
+    #     if "mahlke" in taxonomy.lower():
+    #         taxonomies.mahlke.preprocess(self, resample_params)
+    #
+    #     elif "demeo" in taxonomy.lower():
+    #         taxonomies.demeo.preprocess(self, resample_params)
+    #
+    #     elif "bus" in taxonomy.lower():
+    #         taxonomies.bus.preprocess(self, resample_params)
+    #
+    #     elif "tholen" in taxonomy.lower():
+    #         taxonomies.tholen.preprocess(self, resample_params)
 
-        # Compute latent scores
-        self.data_imputed = model.impute(data_input)
-        self.data_latent = model.transform(self.data_imputed)
+    def classify(self, taxonomy="mahlke", preprocess_remote=False):
+        """Classify a spectrum in a given taxonomic system.
 
-        # Add latent scores and responsibility to input data
-        for factor in range(model.n_factors):
-            input_data[f"z{factor}"] = self.data_latent[:, factor]
+        Parameters
+        ----------
+        taxonomy : str
+            The taxonomic system to use. Choose from ['mahlke', 'demeo', 'bus', 'tholen'].
+            Default is 'mahlke'.
 
-        input_data["cluster"] = np.argmax(self.responsibility, axis=1)
+        Notes
+        -----
+        The classification result is added as 'class_{taxonomy}' attribute to the
+        spectrum instance. Some taxonomies add more than one result as attribute.
+        Refer to the documentation for more information.
+        """
 
-        for i in range(model.n_components):
-            input_data[f"cluster_{i}"] = self.responsibility[:, i]
-
-        # Add asteroid classes based on decision tree
-        self.data_classified = decision_tree.assign_classes(input_data)
-
-        for class_ in defs.CLASSES:
-            setattr(
-                self,
-                f"class_{class_}",
-                self.data_classified[f"class_{class_}"].values[0],
+        # Argument check
+        if taxonomy not in taxonomies.SYSTEMS:
+            raise ValueError(
+                f"Unknown taxonomy '{taxonomy}'. Choose from {taxonomies.SYSTEMS}."
             )
 
-        # Detect features
-        self.data_classified = self.add_feature_flags(self.data_classified)
-        setattr(self, "class_", self.data_classified["class_"].values[0])
+        # Can the spectrum be classified in the requested taxonomy?
+        if not self.is_classifiable(taxonomy):
+            getattr(taxonomies, taxonomy).add_classification_results(self, results=None)
+            return
 
-        # Class per asteroid
-        # self.data_classified = _compute_class_per_asteroid(self.data_classified)
+        # Should spectra from online repositories be preprocessed?
+        # if self.source in sources.SOURCES and preprocess_remote:
+        #     # Get the source-specific preprocessing settings for this taxonomy
+        #     for func, params in (
+        #         getattr(sources, self.source.lower()).PREPROCESSING[taxonomy].items()
+        #     ):
+        #         getattr(self, func)(**params)
 
-        # Print results
-        from collections import Counter
+        # Store for resetting after classification
+        self._wave_pre_class = self.wave.copy()
+        self._refl_pre_class = self.refl.copy()
 
-        classes = Counter(self.data_classified.class_)
-        results_str = ", ".join(
-            [f"{class_}" for class_, count in classes.most_common()]
-        )
-        logger.info(
-            f"[({self.asteroid_number}) {self.asteroid_name}] - [{self.name}]: {results_str}"
+        # Preprocess and classify as defined by scheme
+        getattr(taxonomies, taxonomy).preprocess(self)
+        getattr(taxonomies, taxonomy).classify(self)
+
+        # Reset wavelength and reflectance
+        self.wave = self._wave_pre_class
+        self.refl = self._refl_pre_class
+
+    def is_classifiable(self, taxonomy):
+        """Check if spectrum can be classified in taxonomic scheme based
+        on the wavelength range.
+
+        Parameters
+        ----------
+        taxonomy : str
+            The taxonomic scheme to check.
+
+        Returns
+        -------
+        bool
+            True if the spectrum can be classified, else False.
+        """
+        return getattr(taxonomies, taxonomy.lower()).is_classifiable(self)
+
+    def remove_slope(self, translate_to=None):
+        """Fit a linear function to the spectrum and divide by the fit.
+
+        Parameters
+        ----------
+        translate_to : float
+            Translate the fitted slope to pass trough unity at given wavelength.
+            Useful for DeMeo+ 2009 classification, where the slope should pass
+            through (0.55, 1). Default is None.
+
+        Note
+        ----
+        The fit [slope, intercept] is recorded as ``slope`` attribute.
+        """
+        self.refl, self.slope = preprocessing.remove_slope(
+            self.wave, self.refl, translate_to
         )
 
     def detect_features(self, feature="all", skip_validation=False):
@@ -366,32 +388,32 @@ class Spectrum:
                             )
         return data_classified
 
-    def plot(self, add_classes=False):
-        plotting.plot_spectra([self], add_classes)
+    def plot(self, add_classes=False, taxonomy="mahlke"):
+        plotting.plot_spectra([self], add_classes, taxonomy)
 
-    # def plot(self, show=True):
-    #     """Plot the spectrum.
-    #
-    #     Parameters
-    #     ----------
-    #     show : bool
-    #         Open the plot. Default is True.
-    #
-    #     Returns
-    #     -------
-    #     matplolib.figures.Figure
-    #     matplotlib.axes.Axis
-    #     """
-    #     return plotting._plot_spectrum(self, show)
+    def resample(self, grid, **kwargs):
+        """Resample the spectrum to another wavelength grid.
+
+        Parameters
+        ----------
+        grid : list
+            The target wavelength values.
+
+        Notes
+        -----
+        Any additional parameters are passed to the ``scipy.interpoalte.interp1d`` function.
+        """
+
+        self.refl = preprocessing.resample(self.wave, self.refl, grid, **kwargs)
+        self.wave = grid
 
     def to_csv(self, path_out=None):
         """Store the classification results to file."""
         result = {}
 
         for attr in [
-            "asteroid_name",
-            "asteroid_number",
             "name",
+            "number",
             "class_",
             *[f"class_{letter}" for letter in defs.CLASSES],
         ]:
@@ -408,40 +430,6 @@ class Spectrum:
 
     # ------
     # Utility functions, not to be called directly
-    def _normalize_l2(self):
-        """Normalize the reflectance using the L2 norm."""
-        self.refl = preprocessing.normalize(self.refl.reshape(1, -1))[0]
-
-        if hasattr(self, "refl_original"):
-            self.refl_original = preprocessing.normalize(
-                self.refl_original.reshape(1, -1)
-            )[0]
-
-    def _smooth_interactive(self):
-        """Helper to smooth spectrum interactively. Call the 'smooth' function instead."""
-
-        _, ax = self.plot()
-
-        # Include widgets
-        def update_smooth(_):
-            """Read the GUI values and re-smooth the spectrum."""
-            self.smooth_degree = int(degree_box.text)
-            self.smooth_window = int(window_box.text)
-
-            self.smooth(interactive=False)
-            ax.get_lines()[1].set_ydata(self.refl_smoothed)
-            plt.draw()
-
-        ax_win = plt.axes([0.25, 0.90, 0.05, 0.03])
-        window_box = TextBox(ax_win, "Smoothing Window", initial=self.smooth_window)
-
-        ax_deg = plt.axes([0.45, 0.90, 0.05, 0.03])
-        degree_box = TextBox(ax_deg, "Smoothing Degree", initial=self.smooth_degree)
-
-        window_box.on_submit(lambda x: update_smooth(x))
-        degree_box.on_submit(lambda x: update_smooth(x))
-
-        plt.show()
 
 
 def _basic_checks(wave, refl, unc, flag):
@@ -530,10 +518,18 @@ class Feature:
         xrange = np.arange(0.45, 2.45, 0.0001)
         xrange_fit = (self.lower < xrange) & (self.upper > xrange)
 
-        if (
-            xrange[xrange_fit].min() < wave.min()
-            or xrange[xrange_fit].max() > wave.max()
-        ):
+        # Cut data down to region of interest
+        self.refl = refl[(wave > self.lower - 0.3) & (wave < self.upper + 0.3)]
+        self.wave = wave[(wave > self.lower - 0.3) & (wave < self.upper + 0.3)]
+
+        range_fit = (self.lower < self.wave) & (self.upper > self.wave)
+
+        # if (
+        #     xrange[xrange_fit].min() < wave.min()
+        #     or xrange[xrange_fit].max() > wave.max()
+        #     or len(xrange[xrange_fit]) < 5
+        # ):
+        if len(self.wave[range_fit]) < 4:  # we need at least 4 data points
             logger.debug(
                 f"Passed spectrum does not cover the {self.name}-feature wavelength region."
             )
@@ -543,10 +539,6 @@ class Feature:
         else:
             self.is_observed = True
 
-        # Cut data down to region of interest
-        self.refl = refl[(wave > self.lower - 0.3) & (wave < self.upper + 0.3)]
-        self.wave = wave[(wave > self.lower - 0.3) & (wave < self.upper + 0.3)]
-
         slope = self._fit_continuum()
         refl_no_continuum = self._remove_continuum(slope)
         band = self._fit_band(refl_no_continuum)
@@ -615,7 +607,6 @@ class Feature:
             return True
 
     def remove_slope(self):
-
         # Fit first-order polynomial
         slope = np.polyfit(wave[range_fit], refl[range_fit], 1)
 
@@ -635,7 +626,7 @@ class Spectra(list):
             The name, number, or designation of the asteroid.
         source : str, list of string
             Only return spectrum from specified sources. Choose one or more
-            from classy.data.SOURCES. Default is None, which returns all spectra.
+            from classy.sources.SOURCES. Default is None, which returns all spectra.
         """
 
         if source is not None:
@@ -643,12 +634,12 @@ class Spectra(list):
                 source = [source]
 
             for s in source:
-                if s not in data.SOURCES:
+                if s not in sources.SOURCES:
                     raise ValueError(
-                        f"Unknown source '{s}'. Choose from {data.SOURCES}."
+                        f"Unknown source '{s}'. Choose from {sources.SOURCES}."
                     )
         else:
-            source = data.SOURCES
+            source = sources.SOURCES
 
         # Need this check for __add__
         if not isinstance(id_, list):
@@ -686,28 +677,31 @@ class Spectra(list):
     #         super().extend(str(item) for item in other)
     #
     def __add__(self, rhs):
+        if isinstance(rhs, Spectrum):
+            rhs = [rhs]
         if not isinstance(rhs, list):
-            raise TypeError("Expected a list of classy.Spectrum instances.")
+            raise TypeError(
+                "Expected a list of classy.Spectrum or a single classy.Spectrum."
+            )
         if not all(isinstance(entry, Spectrum) for entry in rhs):
             raise ValueError(
                 "Can only add a list of classy.Spectrum instances or a classy.Spectra instance to another classy.Spectra instance."
             )
         return Spectra([*self, *rhs])
 
-    def plot(self, add_classes=False):
-        plotting.plot_spectra(list(self), add_classes)
+    def plot(self, add_classes=False, taxonomy="mahlke"):
+        plotting.plot_spectra(list(self), add_classes, taxonomy)
 
-    def classify(self):
+    def classify(self, taxonomy="mahlke"):
         for spec in self:
-            spec.classify()
+            spec.classify(taxonomy=taxonomy)
 
     def to_csv(self, path_out=None):
         results = {}
 
         for attr in [
-            "asteroid_name",
-            "asteroid_number",
             "name",
+            "number",
             "class_",
             *[f"class_{letter}" for letter in defs.CLASSES],
         ]:
