@@ -24,6 +24,8 @@ def normalize(spec):
     alpha : float
         The normalization constat to subtract from the spectrum after log-transform.
     """
+    if all(np.isnan(r) for r in spec.refl):
+        return np.nan
 
     # Load the trained normalization instance
     normalization, neighbours = classy.data.load("mixnorm")
@@ -52,9 +54,9 @@ def normalize(spec):
     #         ls="--",
     #         c="red",
     #     )
-    spec.refl_normalised = (
-        spec.refl_interp
-        / np.nanmean(spec.refl_interp)
+    spec.refl= (
+        spec.refl
+        / np.nanmean(spec.refl)
         * np.nanmean(neighbours.loc[idx_nearest_neighbours, classy.defs.WAVE_GRID_STR])
     )
     # plt.plot(
@@ -66,13 +68,12 @@ def normalize(spec):
     # plt.show()
     return alpha
 
-    # Create gamma_init by finidng spectrum closest to the new one
+    # Create gamma_init by finding spectrum closest to the new one
     data_complete = ml_master[classy.ML_SETUP.COLUMNS["spectra"]].values
 
     gamma_init = []
     alpha_init = []
     for ind, spec in data.iterrows():
-
         mask = np.isnan(spec)
         reference = data_complete[:, ~mask]
         reference_normed = reference[~np.isnan(reference).any(axis=1), :]
@@ -189,6 +190,7 @@ def _find_nearest_neighbours(spec, neighbours, N):
     """
 
     # Mask the neighbours to the observed wavelength range
+    spec.mask = np.array([np.isfinite(r) for r in spec.refl])
     neighbours = neighbours[:, spec.mask]
 
     # We cannot compare with neighbours which do not have all bins observed
@@ -197,7 +199,7 @@ def _find_nearest_neighbours(spec, neighbours, N):
     )
 
     # Normalize the spectrum and the neighbours using the L2 norm
-    refl_spec = preprocessing.normalize(spec.refl_interp[spec.mask].reshape(1, -1))
+    refl_spec = preprocessing.normalize(spec.refl[spec.mask].reshape(1, -1))
     neighbours[mask_neighbours] = preprocessing.normalize(neighbours[mask_neighbours])
 
     # Find the closest neighbours in L2 distance
@@ -287,14 +289,12 @@ def gem_mixnorm(
     # GEM algorithm (GM step followed by E step)
 
     while (it <= nit_em) & (reldiff > eps):
-
         it = it + 1
 
         # GM step
         log_pis = np.log(np.sum(gamma, 0)) - np.log(n)  # class proportions
 
         for h in range(nit_ca):  # coordinate ascent loop
-
             Xminusalpha = X - alpha
 
             for k in range(K):
@@ -373,7 +373,6 @@ def mixnorm(
     eps=1e-5,
     verbose=False,
 ):
-
     """Runs the GEM algorithm using multiple random initialisations
 
     INPUT
@@ -494,7 +493,6 @@ def gem_mixnorm_eval(
 
     # GEM algorithm (M step followed by E step)
     while (it <= nit_em) & (reldiff > eps):
-
         it = it + 1
 
         # M step
@@ -547,7 +545,6 @@ def normalize_l2(data):
 
     # handle the missing bins
     for i, sample in enumerate(data_specs):
-
         index = data.index.values[i]
 
         # get incomplete values
