@@ -11,6 +11,7 @@ from matplotlib.widgets import Button, TextBox
 import numpy as np
 
 from classy import cache
+from classy import taxonomies
 
 
 def get_colors(N, cmap="turbo"):
@@ -53,7 +54,6 @@ def plot(data):
     # Plot the distribution in latent space
     for i in range(classy.defs.MODEL_PARAMETERS["d"] - 1):
         for j in range(classy.defs.MODEL_PARAMETERS["d"] - 1):
-
             dx = j  # latent dimension on x-axis
             dy = i + 1  # latent dimension on y-axis
 
@@ -126,49 +126,6 @@ def plot(data):
     plt.show()
 
 
-def _smooth_interactive(spectrum):
-    """Helper to smooth spectrum interactively. Call the 'smooth' function instead."""
-
-    # Suppress unrelated deprecation warning on TextBox.submit
-    warnings.simplefilter("ignore", mpl.MatplotlibDeprecationWarning)
-
-    _, ax = spectrum.plot(show=False)
-
-    # Include widgets
-    def update_smooth(_):
-        """Read the GUI values and re-smooth the spectrum."""
-        spectrum.smooth_degree = int(degree_box.text)
-        spectrum.smooth_window = int(window_box.text)
-
-        spectrum.smooth(interactive=False)
-        ax.get_lines()[1].set_ydata(spectrum.refl_smoothed)
-        plt.draw()
-
-    def update_and_exit(_):
-        """Update the smoothing and quit the plot."""
-        update_smooth(None)
-        plt.close()
-
-    ax_win = plt.axes([0.25, 0.90, 0.05, 0.03])
-    window_box = TextBox(ax_win, "Smoothing Window", initial=spectrum.smooth_window)
-
-    ax_deg = plt.axes([0.45, 0.90, 0.05, 0.03])
-    degree_box = TextBox(ax_deg, "Smoothing Degree", initial=spectrum.smooth_degree)
-
-    ax_update = plt.axes([0.55, 0.90, 0.12, 0.03])
-    update = Button(ax_update, "Update Smoothing")
-
-    ax_exit = plt.axes([0.75, 0.90, 0.12, 0.03])
-    exit_ = Button(ax_exit, "Save and Exit")
-
-    window_box.on_submit(lambda x: update_smooth(x))
-    degree_box.on_submit(lambda x: update_smooth(x))
-    update.on_clicked(lambda x: update_smooth(x))
-    exit_.on_clicked(lambda x: update_and_exit(x))
-
-    plt.show()
-
-
 def _plot_spectrum(spectrum, show=True):
     """Plot the spectrum.
 
@@ -248,7 +205,6 @@ def fit_feature():
     )
 
     def update_lower(lower_limit):
-
         # Update limit value
         FEATURE["lower"] = lower_limit
 
@@ -286,7 +242,6 @@ def fit_feature():
         fig.canvas.draw_idle()
 
     def update_upper(upper_limit):
-
         # Update limit value
         FEATURE["upper"] = upper_limit
 
@@ -331,8 +286,8 @@ def fit_feature():
     plt.show()
 
 
-def plot_spectra(spectra, add_classes=False, system="Mahlke+ 2022"):
-    """Plot spectra. Called by 'classy spectra [id] -p'.
+def plot_spectra(spectra, add_classes=False, taxonomy="mahlke"):
+    """Plot spectra. Called by 'classy spectra [id]'.
 
     Parameters
     ----------
@@ -340,14 +295,19 @@ def plot_spectra(spectra, add_classes=False, system="Mahlke+ 2022"):
         The spectra to plot.
     add_classes : bool
         Add axes showing classification preprocessing and results.
+    taxonomy : str
+        The taxonomic system to plot.
     """
+
+    # Give user some degree of freedom in taxonomy specification
+    taxonomy = taxonomy.lower()
+
     # Ensure uniform plot appearance
     mpl.rcParams.update(mpl.rcParamsDefault)
 
-    # Add color information to spectrum for simplicity
-    # n_smass = sum([1 for spec in spectra if spec.source == "SMASS"])
-    # colors_smass = get_colors(n_smass, "turbo")
+    # Add color information to spectrum
     colors = get_colors(len(spectra), cmap="jet")
+    lines, labels = [], []  # for the global legend
 
     # Build figure instance
     if add_classes:
@@ -356,105 +316,71 @@ def plot_spectra(spectra, add_classes=False, system="Mahlke+ 2022"):
         )
         ax_spec, ax_pv, ax_classes = axes
     else:
-        fig, ax_spec = plt.subplots()
+        fig, ax_spec = plt.subplots(figsize=(12, 7))
 
-    # 1. Plot spectra
-    smass_lines, smass_labels = [], []
-    gaia_lines, gaia_labels = [], []
-    akari_lines, akari_labels = [], []
-    user_lines, user_labels = [], []
-    for spec in spectra:
+    # 1. Plot spectra, grouped by _source
+    _sources = sorted(set(spec._source for spec in spectra))
 
-        if spec.source in ["AKARI", "SMASS", "Gaia"]:
+    for source in _sources:
+        lines_source, labels_source = [], []
+        for spec in spectra:
+            if spec._source != source:
+                continue
 
-            spec.color = colors.pop()
-            if spec.source == "SMASS":
-                smass_line, smass_label = plot_smass_spectrum(ax_spec, spec)
-                smass_lines.append(*smass_line)
-                smass_labels.append(*smass_label)
+            if not hasattr(spec, "wave_plot"):
+                spec.wave_plot = spec.wave
+                spec.refl_plot = spec.refl
 
-            elif spec.source == "Gaia":
-                spec.color = "black"
-                gaia_lines, gaia_labels = plot_gaia_spectrum(ax_spec, spec)
-            elif spec.source == "AKARI":
-                akari_lines, akari_labels = plot_akari_spectrum(ax_spec, spec)
+            spec._color = colors.pop() if source != "Gaia" else "black"
 
-        else:
-            user_line, user_label = plot_user_spectrum(ax_spec, spec)
-            user_lines.append(*user_line)
-            user_labels.append(*user_label)
+            if source == "Gaia":
+                user_line, user_label = plot_gaia_spectrum(ax_spec, spec)
+            else:
+                user_line, user_label = plot_user_spectrum(ax_spec, spec)
+            lines_source.append(*user_line)
+            labels_source.append(*user_label)
 
-    # Axis setup
-    # Construct legend
-    lines, labels = [], []
-
-    if akari_lines:
         (dummy,) = ax_spec.plot([], [], alpha=0)
-        lines += [dummy, dummy] + akari_lines
-        labels += ["", "AKARI"] + akari_labels
-    if gaia_lines:
-        (dummy,) = ax_spec.plot([], [], alpha=0)
-        lines += [dummy, dummy] + gaia_lines
-        labels += ["", "Gaia"] + gaia_labels
-    if smass_lines:
-        (dummy,) = ax_spec.plot([], [], alpha=0)
-        lines += [dummy, dummy] + smass_lines
-        labels += ["", "SMASS"] + smass_labels
-    if user_lines:
-        (dummy,) = ax_spec.plot([], [], alpha=0)
-        lines += [dummy, dummy] + user_lines
-        labels += ["", "User Provided"] + user_labels
+        lines += [dummy, dummy] + lines_source
+        labels += ["", source] + labels_source
+
+    # (dummy,) = ax_spec.plot([], [], alpha=0)
+    # (l1,) = ax_spec.plot([], [], ls="-", c="black", lw=1)
+    # (l2,) = ax_spec.plot([], [], ls="-", c="black", alpha=0.3, lw=3)
+    # lines += [dummy, l1, l2]
+    # labels += ["", "Reflectance", "Uncertainty"]
 
     if add_classes:
+        (l3,) = ax_spec.plot([], [], ls=":", c="gray")
+        lines += [dummy, l3]
+        labels += ["", "Taxonomy Limits"]
 
-        (dummy,) = ax_spec.plot([], [], alpha=0)
-        l1 = ax_spec.errorbar(
-            [], [], yerr=[], capsize=3, ls="", c="black", lw=1, alpha=0.3
-        )
-        (l2,) = ax_spec.plot([], [], ls="-", c="black")
-        lines += [dummy, l1, l2]
-        labels += ["", "Observed", "Preprocessed"]
-    leg = ax_spec.legend(
-        lines,
-        labels,
-        edgecolor="none",
-        loc="center right"
-        # frameon=False,
-        # loc="lower center",
-        # bbox_to_anchor=(1.2, 1.08),
-        # ncol=2,
-    )
+    leg = ax_spec.legend(lines, labels, edgecolor="none", loc="center right")
     ax_spec.add_artist(leg)
-    # if any(spec.source == "SMASS" for spec in spectra):
-    #
-    #     ax_spec.legend(  # SMASS legend, Gaia one is added in function
-    #         facecolor="white",
-    #         edgecolor="none",
-    #         title="SMASS",
-    #         loc="lower right",
-    #     )
 
     if add_classes:
-        ax_spec.axvline(0.45, ls=":", zorder=-10, c="dimgray")
-        ax_spec.axvline(2.45, ls=":", zorder=-10, c="dimgray")
+        wave = getattr(taxonomies, taxonomy).WAVE
+        lower, upper = min(wave), max(wave)
+        ax_spec.axvline(lower, ls=":", zorder=-10, c="gray")
+        ax_spec.axvline(upper, ls=":", zorder=-10, c="gray")
 
     # ensure that there is space for the legend by adding empty space
     xmin, xmax = ax_spec.get_xlim()
     xmax += 1 / 2.8 * (xmax - xmin)
 
-    ax_spec.set(xlabel=r"Wavelength / $\mu$m", ylabel="Reflectance", xlim=(xmin, xmax))
+    ax_spec.set(xlabel=r"Wavelength / µm", ylabel="Reflectance", xlim=(xmin, xmax))
 
     # 2. Add pV axis
     if add_classes:
-
         for i, spec in enumerate(spectra):
             ax_pv.errorbar(
-                i, spec.pV, yerr=spec.pV_err, capsize=3, marker=".", c=spec.color
+                i, spec.pV, yerr=spec.pV_err, capsize=3, marker=".", c=spec._color
             )
 
-        ax_pv.set_xticks(
-            range(len(spectra)), [f"{spec.name}" for spec in spectra], rotation=90
-        )
+        ticklabels = [
+            spec.source if hasattr(spec, "source") else spec._source for spec in spectra
+        ]
+        ax_pv.set_xticks(range(len(spectra)), ticklabels, rotation=90)
 
         ymin, ymax = ax_pv.get_ylim()
         ymin = 0 if ymin < 0.1 else ymin
@@ -464,35 +390,45 @@ def plot_spectra(spectra, add_classes=False, system="Mahlke+ 2022"):
 
     # 3. Add classes
     if add_classes:
+        if "mahlke" in taxonomy:
+            width = 0.8 / len(spectra)
 
-        width = 0.8 / len(spectra)
+            for i, spec in enumerate(spectra):
+                if not hasattr(spec, "class_A"):
+                    continue  # spec was not classified following Mahlke+ 2022
+                for x, class_ in enumerate(classy.defs.CLASSES):
+                    ax_classes.bar(
+                        x - 0.3 + i * width if len(spectra) > 1 else x,
+                        getattr(spec, f"class_{class_}"),
+                        fill=True,
+                        color=spec._color,
+                        width=width,
+                        alpha=0.7,
+                        label=f"{spec.source if hasattr(spec, 'source') else spec._source}: {spec.class_}"
+                        if x == 0
+                        else None,
+                    )
+            ax_classes.set(ylim=(0, 1))
+            ax_classes.set_xticks(
+                [i for i, _ in enumerate(classy.defs.CLASSES)], classy.defs.CLASSES
+            )
+            ax_classes.legend(title="Most Likely Class", frameon=True, edgecolor="none")
+            ax_classes.grid(c="gray", alpha=0.4, zorder=-100)
+        elif "tholen" in taxonomy:
+            ax_classes = taxonomies.tholen.plot_pc_space(ax_classes, spectra)
+        elif "demeo" in taxonomy:
+            ax_classes = taxonomies.demeo.plot_pc_space(ax_classes, spectra)
 
-        for i, spec in enumerate(spectra):
-            if not hasattr(spec, "class_A"):
-                continue  # spec was not classified following Mahlke+ 2022
-            for x, class_ in enumerate(classy.defs.CLASSES):
-                ax_classes.bar(
-                    x - 0.3 + i * width,
-                    getattr(spec, f"class_{class_}"),
-                    fill=True,
-                    color=spec.color,
-                    width=width,
-                    alpha=0.7,
-                    label=f"{spec.name}: {spec.class_}" if x == 0 else None,
-                )
-        ax_classes.set(ylim=(0, 1))
-        ax_classes.set_xticks(
-            [i for i, _ in enumerate(classy.defs.CLASSES)], classy.defs.CLASSES
-        )
-        ax_classes.legend(title="Most Likely Class", frameon=True, edgecolor="none")
-        ax_classes.grid(c="gray", alpha=0.4, zorder=-100)
-
-    if spec.asteroid_name is not None:
-        ax_spec.set_title(
-            f"({spec.asteroid_number}) {spec.asteroid_name}", loc="left", size=10
-        )
+    if spec.name is not None:
+        ax_spec.set_title(f"({spec.number}) {spec.name}", loc="left", size=10)
     if add_classes:
-        ax_classes.set_title(f"Classification following {system}", loc="left", size=10)
+        if taxonomy == "tholen":
+            taxonomy = "Tholen 1984"
+        if taxonomy == "mahlke":
+            taxonomy = "Mahlke+ 2022"
+        ax_classes.set_title(
+            f"Classification following {taxonomy}", loc="left", size=10
+        )
 
     fig.tight_layout()
     plt.show()
@@ -510,37 +446,39 @@ def plot_smass_spectrum(ax, spec):
     """
 
     # Error-interval
-
     if hasattr(spec, "refl_interp"):
         l1 = ax.errorbar(
             spec.wave,
             spec.refl,
             yerr=spec.refl_err,
-            c=spec.color,
-            label=f"{spec.run}",
+            c=spec._color,
             alpha=0.4,
             capsize=3,
             ls="",
         )
-        ax.plot(classy.defs.WAVE_GRID, spec.refl_interp, c=spec.color)
+        ax.plot(spec.wave_interp, spec.refl_interp, c=spec._color)
 
     else:
         # Line
         (l1,) = ax.plot(
-            spec.wave, spec.refl, c=spec.color, label=f"{spec.run}", ls="-", alpha=0.5
+            spec.wave,
+            spec.refl,
+            c=spec._color,
+            ls="-",
+            alpha=0.5,
         )
 
         ax.fill_between(
             spec.wave,
             spec.refl + spec.refl_err / 2,
             spec.refl - spec.refl_err / 2,
-            color=spec.color,
+            color=spec._color,
             alpha=0.3,
             ec="none",
         )
 
     line = [l1]
-    label = [spec.name]
+    label = [spec.shortbib]
     return line, label
 
 
@@ -556,11 +494,17 @@ def plot_gaia_spectrum(ax, spec):
     """
 
     # Line to guide the eye
-    ax.plot(spec.wave, spec.refl, ls=":", lw=1, c=spec.color, zorder=100)
+    ax.plot(spec.wave_plot, spec.refl_plot, ls=":", lw=1, c=spec._color, zorder=100)
 
     # Errorbars colour-coded by photometric flag
     props = dict(lw=1, capsize=3, ls="", zorder=100)
-    l0 = ax.errorbar(spec.wave, spec.refl, yerr=spec.refl_err, c=spec.color, **props)
+    l0 = ax.errorbar(
+        spec.wave,
+        spec.refl,
+        yerr=spec.refl_err,
+        c=spec._color,
+        **props,
+    )
 
     lines = [l0]  # to construct Gaia-specific legend
 
@@ -569,13 +513,21 @@ def plot_gaia_spectrum(ax, spec):
 
     if any(f1):
         l1 = ax.errorbar(
-            spec.wave[f1], spec.refl[f1], yerr=spec.refl_err[f1], c="orange", **props
+            spec.wave[f1],
+            spec.refl[f1],
+            yerr=spec.refl_err[f1],
+            c="orange",
+            **props,
         )
         lines.append(l1)
 
     if any(f2):
         l2 = ax.errorbar(
-            spec.wave[f2], spec.refl[f2], yerr=spec.refl_err[f2], c="red", **props
+            spec.wave[f2],
+            spec.refl[f2],
+            yerr=spec.refl_err[f2],
+            c="red",
+            **props,
         )
         lines.append(l2)
 
@@ -583,11 +535,12 @@ def plot_gaia_spectrum(ax, spec):
 
     if hasattr(spec, "refl_interp"):
         (l3,) = ax.plot(
-            classy.defs.WAVE_GRID,
+            # classy.defs.WAVE_GRID,
+            spec.wave_interp,
             spec.refl_interp,
             ls="-",
             lw=1,
-            c=spec.color,
+            c=spec._color,
             zorder=100,
         )
     # lines.append(l3)
@@ -618,7 +571,7 @@ def plot_akari_spectrum(ax, spec):
     """
 
     # Line to guide the eye
-    ax.plot(spec.wave, spec.refl, ls=":", lw=1, c=spec.color, zorder=100)
+    ax.plot(spec.wave, spec.refl, ls=":", lw=1, c=spec._color, zorder=100)
 
     # Errorbars colour-coded by photometric flag
     props = dict(lw=1, capsize=3, ls="", zorder=100)
@@ -626,7 +579,7 @@ def plot_akari_spectrum(ax, spec):
         spec.wave[spec.flag != 1],
         spec.refl[spec.flag != 1],
         yerr=spec.refl_err[spec.flag != 1],
-        c=spec.color,
+        c=spec._color,
         **props,
     )
 
@@ -639,7 +592,7 @@ def plot_akari_spectrum(ax, spec):
             spec.wave[f1],
             spec.refl[f1],
             yerr=spec.refl_err[f1],
-            c=spec.color,
+            c=spec._color,
             **props,
             alpha=0.3,
         )
@@ -649,11 +602,79 @@ def plot_akari_spectrum(ax, spec):
 
     if hasattr(spec, "refl_interp"):
         (l3,) = ax.plot(
-            classy.defs.WAVE_GRID,
+            spec.wave_interp,
             spec.refl_interp,
             ls="-",
             lw=1,
-            c=spec.color,
+            c=spec._color,
+            zorder=100,
+        )
+    # lines.append(l3)
+
+    # Add Gaia-specific legend
+    # labels += ["Preprocessed"]
+    # leg = ax.legend(
+    #     lines,
+    #     labels,
+    #     facecolor="white",
+    #     edgecolor="none",
+    #     loc="lower center",
+    #     title="Gaia",
+    # )
+    # ax.add_artist(leg)
+    return lines, labels
+
+
+def plot_ecas_spectrum(ax, spec):
+    """Plot an ECAS spectrum.
+
+    Parameters
+    ----------
+    ax : matplotlib.axes.Axis
+        The axis to plot to.
+    spec : classy.spectra.Spectrum
+        The Gaia spectrum to plot.
+    """
+
+    # Line to guide the eye
+    ax.plot(spec.wave, spec.refl, ls="--", lw=1, c=spec._color, zorder=100)
+
+    # Errorbars colour-coded by photometric flag
+    props = dict(lw=1, capsize=3, ls="", zorder=100)
+
+    l0 = ax.errorbar(
+        spec.wave[spec.flag != 1],
+        spec.refl[spec.flag != 1],
+        yerr=spec.refl_err[spec.flag != 1],
+        c=spec._color,
+        **props,
+    )
+
+    lines = [l0]  # to construct AKARI-specific legend
+
+    f1 = spec.flag == 1
+
+    if any(f1):
+        l1 = ax.errorbar(
+            spec.wave[f1],
+            spec.refl[f1],
+            yerr=spec.refl_err[f1],
+            c=spec._color,
+            **props,
+            alpha=0.3,
+        )
+        lines.append(l1)
+
+    labels = [f"Flag {i}" for i, _ in enumerate(lines)]
+
+    if hasattr(spec, "refl_interp"):
+        (l3,) = ax.plot(
+            # classy.defs.WAVE_GRID,
+            spec.wave_interp,
+            spec.refl_interp,
+            ls="-",
+            lw=1,
+            c=spec._color,
             zorder=100,
         )
     # lines.append(l3)
@@ -683,28 +704,26 @@ def plot_user_spectrum(ax, spec):
         The SMASS spectrum to plot.
     """
 
-    # Error-interval
+    # Plot smoothed or original data?
+    if hasattr(spec, "refl_original"):
+        # Plot smoothed
 
-    if hasattr(spec, "refl_interp"):
-        l1 = ax.errorbar(
-            spec.wave,
-            spec.refl,
-            yerr=spec.refl_err,
-            c=spec.color,
-            label=f"{spec.source}",
-            alpha=0.4,
-            capsize=3,
-            ls="",
-        )
-        ax.plot(classy.defs.WAVE_GRID, spec.refl_interp, c=spec.color)
-
-    else:
-        # Line
         (l1,) = ax.plot(
             spec.wave,
             spec.refl,
-            c=spec.color,
-            label=f"{spec.source}",
+            c=spec._color,
+            ls="-",
+            alpha=0.5,
+        )
+
+        ax.plot(spec.wave, spec.refl_original, color=spec._color, alpha=0.3, ls="-")
+    else:
+        # Plot original
+
+        (l1,) = ax.plot(
+            spec.wave,
+            spec.refl,
+            c=spec._color,
             ls="-",
             alpha=0.5,
         )
@@ -713,11 +732,18 @@ def plot_user_spectrum(ax, spec):
             spec.wave,
             spec.refl + spec.refl_err / 2,
             spec.refl - spec.refl_err / 2,
-            color=spec.color,
+            color=spec._color,
             alpha=0.3,
             ec="none",
         )
 
     line = [l1]
-    label = [spec.name]
+
+    if hasattr(spec, "shortbib"):
+        label = spec.shortbib
+    elif hasattr(spec, "source"):
+        label = spec.source
+    else:
+        label = "User"
+    label = [label]
     return line, label
